/* The SpiderMonkey location and history objects implementation. */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "elinks.h"

#include "ecmascript/spidermonkey/util.h"
#include <jsfriendapi.h>

#include "bfu/dialog.h"
#include "cache/cache.h"
#include "cookies/cookies.h"
#include "dialogs/menu.h"
#include "dialogs/status.h"
#include "document/html/frames.h"
#include "document/document.h"
#include "document/forms.h"
#include "document/view.h"
#include "ecmascript/ecmascript.h"
#include "ecmascript/spidermonkey/screen.h"
#include "ecmascript/spidermonkey/window.h"
#include "intl/gettext/libintl.h"
#include "main/select.h"
#include "osdep/newwin.h"
#include "osdep/sysname.h"
#include "protocol/http/http.h"
#include "protocol/uri.h"
#include "session/history.h"
#include "session/location.h"
#include "session/session.h"
#include "session/task.h"
#include "terminal/tab.h"
#include "terminal/terminal.h"
#include "util/conv.h"
#include "util/memory.h"
#include "util/string.h"
#include "viewer/text/draw.h"
#include "viewer/text/form.h"
#include "viewer/text/link.h"
#include "viewer/text/vs.h"

JSClassOps screen_ops = {
	JS_PropertyStub, nullptr,
	JS_PropertyStub, JS_StrictPropertyStub,
	nullptr, nullptr, nullptr, nullptr
};

JSClass screen_class = {
	"screen",
	JSCLASS_HAS_PRIVATE,
	&screen_ops
};

static bool screen_get_property_availHeight(JSContext *ctx, unsigned int argc, JS::Value *vp);
static bool screen_get_property_availWidth(JSContext *ctx, unsigned int argc, JS::Value *vp);
static bool screen_get_property_height(JSContext *ctx, unsigned int argc, JS::Value *vp);
static bool screen_get_property_width(JSContext *ctx, unsigned int argc, JS::Value *vp);

JSPropertySpec screen_props[] = {
	JS_PSG("availHeight",	screen_get_property_availHeight, JSPROP_ENUMERATE),
	JS_PSG("availWidth",	screen_get_property_availWidth, JSPROP_ENUMERATE),
	JS_PSG("height",	screen_get_property_height, JSPROP_ENUMERATE),
	JS_PSG("width",	screen_get_property_width, JSPROP_ENUMERATE),
	JS_PS_END
};

static bool
screen_get_property_availHeight(JSContext *ctx, unsigned int argc, JS::Value *vp)
{
	JS::CallArgs args = CallArgsFromVp(argc, vp);
	JS::RootedObject hobj(ctx, &args.thisv().toObject());

	struct view_state *vs;
	JSCompartment *comp = js::GetContextCompartment(ctx);

	if (!comp) {
		return false;
	}

	struct ecmascript_interpreter *interpreter = JS_GetCompartmentPrivate(comp);

	/* This can be called if @obj if not itself an instance of the
	 * appropriate class but has one in its prototype chain.  Fail
	 * such calls.  */
	if (!JS_InstanceOf(ctx, hobj, &screen_class, NULL))
		return false;

	vs = interpreter->vs;
	if (!vs) {
		return false;
	}

	struct document_view *doc_view = vs->doc_view;

	if (!doc_view) {
		return false;
	}

	args.rval().setInt32(doc_view->box.height * 16);

	return true;
}

static bool
screen_get_property_availWidth(JSContext *ctx, unsigned int argc, JS::Value *vp)
{
	JS::CallArgs args = CallArgsFromVp(argc, vp);
	JS::RootedObject hobj(ctx, &args.thisv().toObject());

	struct view_state *vs;
	JSCompartment *comp = js::GetContextCompartment(ctx);

	if (!comp) {
		return false;
	}

	struct ecmascript_interpreter *interpreter = JS_GetCompartmentPrivate(comp);

	/* This can be called if @obj if not itself an instance of the
	 * appropriate class but has one in its prototype chain.  Fail
	 * such calls.  */
	if (!JS_InstanceOf(ctx, hobj, &screen_class, NULL))
		return false;

	vs = interpreter->vs;
	if (!vs) {
		return false;
	}

	struct document_view *doc_view = vs->doc_view;

	if (!doc_view) {
		return false;
	}

	args.rval().setInt32(doc_view->box.width * 8);

	return true;
}

static bool
screen_get_property_height(JSContext *ctx, unsigned int argc, JS::Value *vp)
{
	JS::CallArgs args = CallArgsFromVp(argc, vp);
	JS::RootedObject hobj(ctx, &args.thisv().toObject());

	struct view_state *vs;
	JSCompartment *comp = js::GetContextCompartment(ctx);

	if (!comp) {
		return false;
	}

	struct ecmascript_interpreter *interpreter = JS_GetCompartmentPrivate(comp);

	/* This can be called if @obj if not itself an instance of the
	 * appropriate class but has one in its prototype chain.  Fail
	 * such calls.  */
	if (!JS_InstanceOf(ctx, hobj, &screen_class, NULL))
		return false;

	vs = interpreter->vs;
	if (!vs) {
		return false;
	}

	struct document_view *doc_view = vs->doc_view;

	if (!doc_view) {
		return false;
	}

	struct session *ses = doc_view->session;

	if (!ses) {
		return false;
	}

	args.rval().setInt32(ses->tab->term->height * 16);

	return true;
}

static bool
screen_get_property_width(JSContext *ctx, unsigned int argc, JS::Value *vp)
{
	JS::CallArgs args = CallArgsFromVp(argc, vp);
	JS::RootedObject hobj(ctx, &args.thisv().toObject());

	struct view_state *vs;
	JSCompartment *comp = js::GetContextCompartment(ctx);

	if (!comp) {
		return false;
	}

	struct ecmascript_interpreter *interpreter = JS_GetCompartmentPrivate(comp);

	/* This can be called if @obj if not itself an instance of the
	 * appropriate class but has one in its prototype chain.  Fail
	 * such calls.  */
	if (!JS_InstanceOf(ctx, hobj, &screen_class, NULL))
		return false;

	vs = interpreter->vs;
	if (!vs) {
		return false;
	}

	struct document_view *doc_view = vs->doc_view;

	if (!doc_view) {
		return false;
	}

	struct session *ses = doc_view->session;

	if (!ses) {
		return false;
	}

	args.rval().setInt32(ses->tab->term->width * 8);

	return true;
}
